/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_mesh.c: triangle model functions

#include "quakedef.h"


unsigned int r_meshindexbuffer = 0;
unsigned int r_meshvertexbuffer = 0;

void GL_MakeAliasModelDisplayLists (stvert_t *stverts, dtriangle_t *triangles)
{
	int i, j;
	aliasmesh_t *hunkdesc;

	// there can never be more than this number of verts
	aliasmesh_t *desc = (aliasmesh_t *) malloc (sizeof (aliasmesh_t) * pheader->numverts);

	// there will always be this number of indexes
	unsigned short *indexes = (unsigned short *) Hunk_Alloc (sizeof (unsigned short) * pheader->numtris * 3);

	// johnfitz -- padded skins
	float hscale = (float) pheader->skinwidth / (float) TM_PadConditional (pheader->skinwidth);
	float vscale = (float) pheader->skinheight / (float) TM_PadConditional (pheader->skinheight);
	// johnfitz

	pheader->indexes = (int) indexes - (int) pheader;
	pheader->numindexes = 0;
	pheader->numverts = 0;

	for (i = 0; i < pheader->numtris; i++)
	{
		for (j = 0; j < 3; j++)
		{
			int v;

			// index into hdr->vertexes
			unsigned short vertindex = triangles[i].vertindex[j];

			// basic s/t coords
			int s = stverts[vertindex].s;
			int t = stverts[vertindex].t;

			// check for back side and adjust texcoord s
			if (!triangles[i].facesfront && stverts[vertindex].onseam) s += pheader->skinwidth / 2;

			// see does this vert already exist
			for (v = 0; v < pheader->numverts; v++)
			{
				// it could use the same xyz but have different s and t
				if (desc[v].vertindex == vertindex && (int) desc[v].st[0] == s && (int) desc[v].st[1] == t)
				{
					// exists; emit an index for it
					indexes[pheader->numindexes++] = v;

					// no need to check any more
					break;
				}
			}

			if (v == pheader->numverts)
			{
				// doesn't exist; emit a new vert and index
				indexes[pheader->numindexes++] = pheader->numverts;

				desc[pheader->numverts].vertindex = vertindex;
				desc[pheader->numverts].st[0] = s;
				desc[pheader->numverts++].st[1] = t;
			}
		}
	}

	// create a hunk buffer for the final mesh we'll actually use
	hunkdesc = (aliasmesh_t *) Hunk_Alloc (sizeof (aliasmesh_t) * pheader->numverts);
	pheader->aliasmesh = (int) hunkdesc - (int) pheader;

	// tidy up the verts by calculating final s and t and copying out to the hunk
	for (i = 0; i < pheader->numverts; i++)
	{
		hunkdesc[i].vertindex = desc[i].vertindex;
		hunkdesc[i].st[0] = (((float) desc[i].st[0] + 0.5f) / (float) pheader->skinwidth) * hscale;
		hunkdesc[i].st[1] = (((float) desc[i].st[1] + 0.5f) / (float) pheader->skinheight) * vscale;
	}

	// don't forget!!!
	free (desc);
}


void R_CreateMeshIndexBuffer (void)
{
	if (gl_arb_vertex_buffer_object)
	{
		int i;
		int r_nummeshindexes = 0;
		int r_nummeshvertexes = 0;

		if (r_meshindexbuffer)
		{
			qglDeleteBuffers (1, &r_meshindexbuffer);
			r_meshindexbuffer = 0;
		}

		if (r_meshvertexbuffer)
		{
			qglDeleteBuffers (1, &r_meshvertexbuffer);
			r_meshvertexbuffer = 0;
		}

		// count the vertexes and indexes used
		// this can't be done at load time owing to the caching system meaning that some models may not be loaded for a new map!!!
		r_nummeshindexes = 0;
		r_nummeshvertexes = 0;

		// fill in indexes for the alias models
		for (i = 1; i < MAX_MODELS; i++)
		{
			model_t *m;
			aliashdr_t *hdr;

			if (!(m = cl.model_precache[i])) break;
			if (m->type != mod_alias) continue;

			hdr = (aliashdr_t *) Mod_Extradata (m);

			r_nummeshindexes += hdr->numindexes;
			r_nummeshvertexes += hdr->numverts;
		}

		qglGenBuffers (1, &r_meshindexbuffer);
		qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, r_meshindexbuffer);
		qglBufferData (GL_ELEMENT_ARRAY_BUFFER, r_nummeshindexes * sizeof (unsigned short), NULL, GL_STATIC_DRAW);

		qglGenBuffers (1, &r_meshvertexbuffer);
		qglBindBuffer (GL_ARRAY_BUFFER, r_meshvertexbuffer);
		qglBufferData (GL_ARRAY_BUFFER, r_nummeshvertexes * sizeof (float) * 2, NULL, GL_STATIC_DRAW);

		// begin counting again for offsets
		r_nummeshindexes = 0;
		r_nummeshvertexes = 0;

		// fill in indexes for the alias models
		for (i = 1; i < MAX_MODELS; i++)
		{
			int j;
			model_t *m;
			aliashdr_t *hdr;
			unsigned short *ndx;
			aliasmesh_t *mesh;
			float *st, *dstst;

			if (!(m = cl.model_precache[i])) break;
			if (m->type != mod_alias) continue;

			hdr = (aliashdr_t *) Mod_Extradata (m);
			mesh = (aliasmesh_t *) ((byte *) hdr + hdr->aliasmesh);

			// this stores the offset to the index, not the index of it (fixme; rename it)
			hdr->firstindex = r_nummeshindexes * sizeof (unsigned short);
			r_nummeshindexes += hdr->numindexes;

			hdr->firstvertex = r_nummeshvertexes * sizeof (float) * 2;
			r_nummeshvertexes += hdr->numverts;

			qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, r_meshindexbuffer);
			ndx = (unsigned short *) ((byte *) hdr + hdr->indexes);
			qglBufferSubData (GL_ELEMENT_ARRAY_BUFFER, hdr->firstindex, hdr->numindexes * sizeof (unsigned short), ndx);

			st = (float *) malloc (hdr->numverts * sizeof (float) * 2);
			dstst = st;

			for (j = 0; j < hdr->numverts; j++, mesh++, dstst += 2)
			{
				dstst[0] = mesh->st[0];
				dstst[1] = mesh->st[1];
			}

			qglBindBuffer (GL_ARRAY_BUFFER, r_meshvertexbuffer);
			qglBufferSubData (GL_ARRAY_BUFFER, hdr->firstvertex, hdr->numverts * sizeof (float) * 2, st);
			free (st);
		}

		// unbind the buffer when done
		qglBindBuffer (GL_ELEMENT_ARRAY_BUFFER, 0);
		qglBindBuffer (GL_ARRAY_BUFFER, 0);

		// for the next map
		r_nummeshindexes = 0;
		r_nummeshvertexes = 0;
	}
}
